import  random, operator, math, spherogram, numpy
class rBraid(): 
    ""
    def __init__(self,n,k,filename=""):
      	# Author: Mauricio B. Garcia Tec
        # Date: February 2014
        # In:
        #     n.- the number of strands in the braid group.
        #     k.- the length of the random walk.
        # Out:
        #     A knot or link which is the closure of the random braid   
        random.seed()

        const = max(1,k/25.0) # a posteriori resizes the knot conveniently, controls the height/width ratio.
        
        # Bottom of the braid.
        strand_pos = [range(n)]+[range(n)] # contains at which vertex the position of the i-th strand.
        swap, swapinv = range(n), range(n) # helpful during the walk.
        height = [[ (s+1) for s in range(n)]] + [[n+1 for s in range(n)]] 
        crossings = [] # Pairs indicating the positions of the n crossings. Left edge passes under the right edge.
        cross_sph_list = [ [] for s in range(n)] # This is the memory intensive part. We will create a list of length n. Each elements is a sublist
                           # of triples [a,b,c] where a E {0,1,2,...,n-1} indicates the vertices touched by each strand and 
                           # and b E {-1,1} indicates if the flow of the strand at that vertex is above (1) or below (-1).
        cross_orientation = []  # vector with values {0,1} indication type of crossing orientation, right over left (0), otherwise (1)
        
        # The random walk.
        for i in range(k): # the random walk.
            gen = random.randint(1,2*(n-1)) # each number represents with equal probability a generator of the braid group.
                                            # each pair is a generator and its inverse.
            pos = (gen-1)/2 # connects the pos-th and (pos+1)-th strands.
            above = (gen-1)%2 # tells which strand passes above. If 1 then left over right, 0 otherwise.
            height += [[(n + 2 + i) for s in range(n)]] # for simplicity in the resulting code, we will generate 2*k levels during the walk.
            ### strand position (currently pretty unefficient).
            swap_aux = swap[pos]
            swap[pos] = swap[pos+1]
            swap[pos+1] = swap_aux
            swap_inv = [[swap[j],j] for j in range(n)]
            swap_inv.sort(key=operator.itemgetter(0))
            swap_inv = [row[1] for row in swap_inv]
            ### 
            strand_pos += [[s for s in swap_inv]]
            cross1 = swap[pos]
            cross2 = swap[pos+1]
            if above == 0:
                crossings += [[cross1,cross2]]
                cross_sph_list[swap[pos]] += [[i,-1]]
                cross_sph_list[swap[pos+1]] += [[i,1]]
            else:
                crossings += [[cross2,cross1]]
                cross_sph_list[swap[pos]] += [[i,1]]
                cross_sph_list[swap[pos+1]] += [[i,-1]]
            cross_orientation += [above]

        # Top of the braid.
        strand_pos += [swap_inv] 
        height += [[(2*n+k+1-swap_inv[s]) for s in range(n)]]      
            
        # Detect the number of components factoring the las permutation. Each cycle stands for a component.
        remaining = {s for s in range(n)}
        perm = [swap_inv[i] for i in range(n)]
        cycles = [] 
        subcycle = []
        place,smallest = 0,0
        while remaining!=set():
            if place in remaining:
                remaining.remove(place)
                auxplace = place
                place = perm[auxplace]
                subcycle += [place]
            else:
                cycles += [[s for s in subcycle]]
                smallest += 1
                while not smallest in remaining and smallest < n-1:
                    smallest += 1                    
                place = smallest
                subcycle = []
        cycles += [[s for s in subcycle]] # is the cycle factorization of the permutation.
        linkcomponents = len(cycles)
        self.num_components = linkcomponents
        
        # Note on vertex counting: to each strand we will add 2 points to close the strand. Each strand has
        #   k+3 vertices as we defined them, that makes k+5 vertices.
            
        # The variable output will be use write the link in a .lnk file.
        output = '% Link Projection \n'+str(linkcomponents)+'\n'
        output += '   0 0\n'
        cycle_pos = [len(cycles[0])]
        
        for t in range(linkcomponents-1): 
            next = t+1
            cycle_pos += [cycle_pos[t] + len(cycles[next])]
            
        for comp in range(linkcomponents-1): # link component closure and opening.
            output += '   '+str(cycle_pos[comp]*(k+5))+' '+str(cycle_pos[comp]*(k+5))+'\n'
                
        # Vertex printing            
        vertexnumber = n*(k+5)
        output += str(vertexnumber)+'\n'
        perm = [s for sublist in cycles for s in sublist] # we print vertices in cycle order, in reverse order as swap.
        
        for i in range(n):
            x = [const*row[perm[i]] for row in strand_pos]
            y = [row[perm[i]] for row in height]
            lastx = x[k+2]/const # just for better visualizing
            # ---- now we add two points for closing the braid (try to picture this).
            x += [2*const*n-lastx] + [2*const*n-lastx] # arbitrarily moving the points to the right to close them.
            y += [2*n+k+1-lastx] + [lastx+1]
            for j in range(k+5):
                output += '   '+str(x[j])+' '+str(-y[j])+'\n'
                    
        # Edges printing            
        output += str(vertexnumber)+'\n'
        cycle_pos = [0] + [(k+5)*s for s in cycle_pos] 
        
        for s in range(linkcomponents):
            for t in range(cycle_pos[s],cycle_pos[s+1]):
                output += '   '+str(t)+' '+str((t+1)%cycle_pos[s+1]+cycle_pos[s]*((t+1)/cycle_pos[s+1]))+'\n'
                    
        # Crossings printing
        output += str(k)+'\n'
        # NOTE on crossings: luckily we know that there are exactly k crossings and where they take place.         
        perm_inv = [[perm[i],i] for i in range(n)]
        perm_inv.sort(key=operator.itemgetter(0))
        perm_inv = [row[1] for row in perm_inv] 
        # invert perm to know rewerite crossings in terms of edge numbers. For example if 
        # there is a cross between 2 and 1 but 2 was printed first, then we want pinv so that pinv(2)=1.    
        crossings = [[(k+5)*perm_inv[s[0]],(k+5)*perm_inv[s[1]]] for s in crossings]
        crossings = [[crossings[i][0]+i+1,crossings[i][1]+i+1] for i in range(k)]

        for s in range(k):
            output += '   '+str(crossings[s][0])+' '+str(crossings[s][1])+'\n'

        output += str(-1)	
    
        if (filename!=""):
            namelen = len(filename)
            if (namelen >4):
                if (filename[(namelen-4):namelen]!=".lnk"):
                    filename = filename + ".lnk"
            else:
                filename = filename + ".lnk"
            self.filename = filename
            with open(filename,'wt') as file_output:
                file_output.write(output)
            print('A file named "'+filename+'" was created in the current working directory.')

        # For the spherogram module.
        Sph_Crossings = [spherogram.links.links.Crossing(j) for j in range(k)]
        # we will glue cross_sph_list according to observed cycles.
        for i in range(linkcomponents):
            current_cycle = cycles[i]
            len_current_cycle = len(current_cycle)
            cross_list = []
            for s in range(len_current_cycle):
                cross_list += cross_sph_list[current_cycle[s]]
            N = len(cross_list)
            for j in range(N): # we have to proceed by cases.
                vertex = cross_list[j]
                if vertex[1] == -1: # if present vertex flows above or below.
                    prev = 0
                    proc = 2
                else:
                    if cross_orientation[vertex[0]] == 0: # left over right type
                        prev = 3
                        proc = 1
                    else:
                        prev = 1
                        proc = 3
                # 1.- Connect to previous in cycle.
                prev_vertex = cross_list[(j-1)%N]
                if prev_vertex[1] == -1: # if previous flows below, it is easy, it connects to 2.
                    Sph_Crossings[vertex[0]][prev] = Sph_Crossings[prev_vertex[0]][2]
                else: # we have to detect orientation of previous.
                    if cross_orientation[prev_vertex[0]] == 0: # previous is left over right
                        Sph_Crossings[vertex[0]][prev] = Sph_Crossings[prev_vertex[0]][1]
                    else:
                        Sph_Crossings[vertex[0]][prev] = Sph_Crossings[prev_vertex[0]][3]
                # 2.- Connect to proceeding in cycle.
                next_vertex = cross_list[(j+1)%N]
                if  next_vertex[1] == -1: # if proceeding flows below, it is easy, it connects to 2.
                    Sph_Crossings[vertex[0]][proc] = Sph_Crossings[next_vertex[0]][0]
                else: # we have to detect orientation of previous.
                    if cross_orientation[next_vertex[0]] == 0: # proceeding is right over left
                        Sph_Crossings[vertex[0]][proc] = Sph_Crossings[next_vertex[0]][3]
                    else:
                        Sph_Crossings[vertex[0]][proc] = Sph_Crossings[next_vertex[0]][1]                
                            
        # Now we craft the link.
        self.manifold = spherogram.links.links.Link(Sph_Crossings).exterior()



from setuptools import setup

setup(name='rManifolds',
      version='0.1',
      description='Manifolds in snappy from random knots/links/braids',
      url='http://github.com/mauriciogtec/rManifolds',
      author='Mauricio B. Garcia Tec',
      license='MIT',
      packages=['rManifolds'],
      zip_safe=False)



import  random, operator, math, spherogram, numpy
class gaussian_walk(): 
    ""
    def __init__(self, n, filename="", sigma=1):
    	# Author: Mauricio B. Garcia Tec.
    	# Date: Februrary 2014.
    	# This function generates a random knot in SnapPy using the module Spherogram

        # In: int n.-the number of desired random vertices.
        # Out: a object of type Link in SnapPy

        # Generation of n random points following a gaussian distribution and given that X_0=X_(n+1)=0
        x,y,z = [0],[0],[0]        
        for i in range(1,n):
            cond_sd = sigma*math.sqrt((n-i)/(1.0*(n+1-i))) 
            # Random walk given that X_n = 0. Conditional variance is (s/t)*(t-s) where t is the time at 
            # the known future state and current time. By past independence, we can set t = n+1-i and s = 1
            cond_mean_x = -x[i-1]/(1.0*(n+1-i))
            cond_mean_y = -y[i-1]/(1.0*(n+1-i))
            cond_mean_z = -x[i-1]/(1.0*(n+1-i))
            x += [x[i-1] + numpy.random.normal(cond_mean_x,cond_sd,1)[0]]	
            y += [y[i-1] + numpy.random.normal(cond_mean_y,cond_sd,1)[0]]
            z += [z[i-1] + numpy.random.normal(cond_mean_z,cond_sd,1)[0]]

        # The rest of the algorithm is the same as in the uniform case. I will code comments.
        output = '% Link Projection \n1 \n   0    0'	  
        output += '\n'+str(n)
        for i in range(n):
            output += '\n   '+str(x[i])+'  '+str(y[i])
        output += '\n'+str(n)
        for i in range(n):
            output += '\n    '+ str((i+1) % n) +'  '+str(i)
        crossinfo = [] 
        crossid = 0 
        under = [] 
        above = []  
        for i in range(n):
            a0,b0,c0,a1,b1,c1 = x[i],y[i],z[i],x[(i+1)%n],y[(i+1)%n],z[(i+1)%n] 
            for j in range(i-1):

                x0,y0,z0,x1,y1,z1 = x[j],y[j],z[j],x[(j+1)%n],y[(j+1)%n],z[(j+1)%n]  
                A,B,C,D = a1-a0,-(x1-x0),b1-b0,-(y1-y0)
                det = A*D-B*C
                t = (D*(x0-a0)-B*(y0-b0))/det
                u = (-C*(x0-a0)+A*(y0-b0))/det	
                if 0<t and t<1 and 0<u and u<1:			
                    if c0+t*(c1-c0) > z0+u*(z1-z0):
                        aux = 1 
                        under += [j]
                        above += [i]
                    else: 	
                        aux = -1
                        under += [i]
                        above += [j]
                    coords = [x0+u*(x1-x0),y0+u*(y1-y0),z0+u*(z1-z0)] 
                    signum = math.copysign(1,(x1-x0)*(b1-b0)-(a1-a0)*(y1-y0))
                    crossinfo += [(j,i,aux,u,aux*signum,coords,crossid)] 
                    crossinfo += [(i,j,-aux,t,aux*signum,coords,crossid)] 
                    crossid += 1 		
        N = len(crossinfo) 
        totcross = N/2 	
        crossinfo.sort(key=operator.itemgetter(0,3)) 
        crossid = [row[6] for row in crossinfo] 

        output += '\n'+str(totcross)
        for i in range(totcross):
            output += '\n   '+str(under[i])+'    '+str(above[i])
        output += '\n'+str(-1)	
        if (filename!=""):
            namelen = len(filename)
            if (namelen >4):
                if (filename[(namelen-4):namelen]!=".lnk"):
                    filename = filename + ".lnk"
            else:
                filename = filename + ".lnk"
            with open(filename,'wt') as file_output:
                file_output.write(output)
	    self.filename = filename
            print('A file named "'+filename+'" was created in the current working directory.')
    
        crossings = [spherogram.links.links.Crossing(j) for j in range(totcross)] # to start a variable for crossing conventions
        k = 0
        while k < N:
            if crossinfo[k][2]==-1:
                prev = 0
                proc = 2
            else: 
                if crossinfo[k][4]==1:
                    prev = 3			
                    proc = 1
                else:
                    prev = 1
                    proc = 3
            if crossinfo[(k-1)%N][2]==-1: 
                crossings[crossid[k]][prev] = crossings[crossid[(k-1)%N]][2] 		
            else: 
                if crossinfo[(k-1)%N][4] == 1:
                    crossings[crossid[k]][prev] = crossings[crossid[(k-1)%N]][1] 
                else:
                    crossings[crossid[k]][prev] = crossings[crossid[(k-1)%N]][3]
            if crossinfo[(k+1)%N][2]==-1: 
                crossings[crossid[k]][proc] = crossings[crossid[(k+1)%N]][0] 		
            else: 
                if crossinfo[(k+1)%N][4] == 1: 
                    crossings[crossid[k]][proc] = crossings[crossid[(k+1)%N]][3] 
                else:
                    crossings[crossid[k]][proc] = crossings[crossid[(k+1)%N]][1]						
            k += 1	
        self.manifold = spherogram.links.links.Link(crossings).exterior()



import  random, operator, math, spherogram, numpy
class rUnifKnot(): 
    ""
    def __init__(self,n,filename=""):
    	# Author: Mauricio B. Garcia Tec.
    	# Date: Februrary 2014.
    	# This function generates a random knot in SnapPy using the module Spherogram

        # In: int n.-the number of desired random vertices.
        # Out: a object of type Link in SnapPy
        random.seed() # iniatilize random number generator
        self.num_components = 1
    
        # Generation of n random vertices.
        const = 500 # So that the knot can be visualized in the PLink editor witout automatic scaling
        x = [const*random.random() for i in range(n)]	
        y = [const*random.random() for i in range(n)]
        z = [const*random.random() for i in range(n)]
    
        # 1. This information shows that a knot is to be saved (a one-component link).
        output = '% Link Projection \n1 \n   0    0'	
    
        # 2. Prints vertices information.
        output += '\n'+str(n)
        for i in range(n):
            output += '\n   '+str(x[i])+'  '+str(y[i])
        # 3. Connectivity information in standard form.
        output += '\n'+str(n)
        for i in range(n):
            output += '\n    '+ str((i+1) % n) +'  '+str(i)
        
        # 1. Crossings: 
        # CHALLENGE: change from vertex information to crossings information.
        # We shall add vertex by vertex, looking for crossings with the ones already added.
        crossinfo = [] # used for storing crossing information. A list with 6-tuples containing
		       # (a) an edge, (b) another edge (each crossing pair appears twice)
		       # (c) 1 if first edge goes above, -1 if second index goes above
		       # (d) relative distance from the crossing point to the origin of the first-edge
		       # (e) orientation of the crossing: 1 if the edge passing above is going to the right else -1
		       # (f) (pot) the coordinates of the point
		       # (g) crossid (see below)
        crossid = 0 # an id for crossings since each crossing appears twice.
    
        # The following two will be used when printing the .lnk file
        under = [] # 
        above = [] # 
    
        for i in range(n):
            # We add the i-th edge and then check for crossings with the ones already added. We use %n% since it is a circular list.
            a0,b0,c0,a1,b1,c1 = x[i],y[i],z[i],x[(i+1)%n],y[(i+1)%n],z[(i+1)%n] 
            for j in range(i-1):
                # Projection onto the x an y coordinates and use a parametric model to verify there is a crossing.
                # The formula in use comes from solving the system of two linear equations on t and u:  
                #	                (x0,y0)+t(x1-x0,y1-y0) = (a0,b0)+u(a1-a0,b1-b0).
                # If 0 <= t,u <=1, then there is a crossing between two line segements.
                x0,y0,z0,x1,y1,z1 = x[j],y[j],z[j],x[(j+1)%n],y[(j+1)%n],z[(j+1)%n]  
                A,B,C,D = a1-a0,-(x1-x0),b1-b0,-(y1-y0)
                det = A*D-B*C
                t = (D*(x0-a0)-B*(y0-b0))/det
                u = (-C*(x0-a0)+A*(y0-b0))/det	
                if 0<t and t<1 and 0<u and u<1:			
                    if c0+t*(c1-c0) > z0+u*(z1-z0):
                        aux = 1 # first index goes above.
                        under += [j]
                        above += [i]
                    else: 	
                        aux = -1 # second-index goes above.
                        under += [i]
                        above += [j]
                    # To know whether the edge above is going to the right or to the left with respect to edge below
                    # we will use the determinant.
                    # convention: 1 = the above edge is to the right with respect to the edge below	
                    #            -1 = the above edge is to the left with respect to the edge below 
                    coords = [x0+u*(x1-x0),y0+u*(y1-y0),z0+u*(z1-z0)] # crossing coordinates (if crossing)
                    signum = math.copysign(1,(x1-x0)*(b1-b0)-(a1-a0)*(y1-y0))
                    crossinfo += [(j,i,aux,u,aux*signum,coords,crossid)] # the smaller the u, the closest it is to the vertex vec[j]
                    crossinfo += [(i,j,-aux,t,aux*signum,coords,crossid)] # the smaller the t, the closest it is to the vertex vec[i]
                    # Saving the t and u allows us to recreate the crossing information while going around the knot.
                    crossid += 1 # one more crossing		
        N = len(crossinfo) # vector length.
        totcross = N/2 # total number of crossings found.	
        # We want to sort the crossings as if we were traveling the from one vertex to the next in order and going around the knot.
        crossinfo.sort(key=operator.itemgetter(0,3)) #  (Nice stuff!).
        crossid = [row[6] for row in crossinfo] # we extract the crossid column for simplicit
    
        # Print the .lnk file for the PLink editor.
        output += '\n'+str(totcross)
        for i in range(totcross):
            output += '\n   '+str(under[i])+'    '+str(above[i])
        output += '\n'+str(-1)	
        if (filename!=""):
            namelen = len(filename)
            if (namelen >4):
                if (filename[(namelen-4):namelen]!=".lnk"):
                    filename = filename + ".lnk"
            else:
                filename = filename + ".lnk"
            self.filename = filename
            with open(filename,'wt') as file_output:
                file_output.write(output)
            print('A file named "'+filename+'" was created in the current working directory.')
    
        # Use the Spherogram module instead.
        crossings = [spherogram.links.links.Crossing(j) for j in range(totcross)] # to start a variable for crossing conventions
        # We need to mark each crossing
        k = 0
        while k < N:
		# We will proceed as follows:
		# For each crossing in visit order (crossinfo) we will have two cases: whether the flows is going above or below.
		# In both cases we look at the previous and the following knot and with that determine how to tie the crossings.
		# The convention is that each crossing has entries 0,1,2,3 where 0 corresponds to the edge incoming from below
		# having the lowest index. Entries 1,2,3 and assigned moving counterclockwise.
		# There will be 3 interesting cases. If the flow goes below, if the flow goes above in standard direction (the above
		# edge passes to the right) and if it goes above in skew direction (to the left).
		# FIRST CASE	
		# We know in this case that the previous crossing in the list will be connected to  crossings[k][0] and 
		# the one that follows to crossings[k][2]. We need now to check the crossing type of the previous and
		# the following to know to which of their entries to connect.
            if crossinfo[k][2]==-1:
                prev = 0
                proc = 2
                # SECOND CASE	
                # If above then the sign will determine which entry to connect.
            else: 
                if crossinfo[k][4]==1:
                    prev = 3			
                    proc = 1
                    # THIRD CASE
                else:
                    prev = 1
                    proc = 3
            # Now we connect to the previous and next crossings again depending on their type.
            if crossinfo[(k-1)%N][2]==-1: # Case a) The previous crossing's flow is below.
                crossings[crossid[k]][prev] = crossings[crossid[(k-1)%N]][2] 		
            else: # Case b) If the flow is going above, we need to know we must look at the signum (left or right).
                if crossinfo[(k-1)%N][4] == 1: #  Case i) standard direction, the above edge passes to the right.
                    crossings[crossid[k]][prev] = crossings[crossid[(k-1)%N]][1] 
                else: # Case ii) skew direction
                    crossings[crossid[k]][prev] = crossings[crossid[(k-1)%N]][3]
            # The instructions are similar for the following cross.
            if crossinfo[(k+1)%N][2]==-1: # Case a) The follwing crossing's flow is below.
                crossings[crossid[k]][proc] = crossings[crossid[(k+1)%N]][0] 		
            else: # Case b) Flow goes above, we look at signum
                if crossinfo[(k+1)%N][4] == 1: # Case i) standard direction, the above edge passes to the right.
                    crossings[crossid[k]][proc] = crossings[crossid[(k+1)%N]][3] 
                else: # Case ii) skew direction
                    crossings[crossid[k]][proc] = crossings[crossid[(k+1)%N]][1]						
            k += 1	
        # Now we craft the knot.
        self.manifold = spherogram.links.links.Link(crossings).exterior()



from .rBraid import *
from .rUnifKnot import *
from .gaussian_walk import *

try:
    import snappy
except ImportError:
    pass

# def __init__(self):
#	__all__ = ["rBraid", "rUnifKnot", "Gaussian_walk"]
